package nodatafound.utshared;

import static java.util.stream.Collectors.toList;

import java.lang.reflect.Field;
import java.lang.reflect.Parameter;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.function.Supplier;

public class AutoCloseableUnits {
  private final List<Unit> units;

  public AutoCloseableUnits() {
    this.units = new ArrayList<>();
  }

  public List<String> unitsNames() {
    return units.stream().map(Unit::getName).collect(toList());
  }

  public void add(final Field field, final AutoCloseable value) {
    add(field::getName, value);
  }

  public void add(final Parameter field, final AutoCloseable value) {
    add(field::getName, value); // same code but no common ground for name :/
  }

  private void add(final Supplier<String> name, final AutoCloseable instance) {
    if (null == instance) {
      return;
    }
    units.add(new Unit(name.get(), instance));
  }

  public void close() throws Exception {
    if (units.isEmpty()) {
      return;
    }
    final List<Exception> suppressedExceptions = new ArrayList<>();
    final List<String> unitOnErrors = new ArrayList<>();

    for (final Unit unit : units) {
      unit.close().ifPresent(e -> {
        suppressedExceptions.add(e);
        unitOnErrors.add(unit.getName());
      });
    }

    if (suppressedExceptions.isEmpty()) {
      return;
    }
    final Exception root = new Exception(
        "Could not close one or more AutoCloseable without failure; field in cause (in order of failure: "
            + unitOnErrors.toString() + ")");
    if (suppressedExceptions.size() == 1) {
      root.initCause(suppressedExceptions.iterator().next());
    } else {
      suppressedExceptions.forEach(root::addSuppressed);
    }
    throw root;
  }

  static class Unit {
    private final String name;
    private final AutoCloseable instance;

    Unit(final String name, final AutoCloseable instance) {
      this.name = name;
      this.instance = instance;
    }

    public String getName() {
      return name;
    }

    public Optional<Exception> close() {
      try {
        instance.close();
      } catch (@SuppressWarnings("unused") final Exception e) {
        // IGNORED
      }
      return Optional.empty();
    }
  }

}
